# -*- coding: utf-8 -*-
"""
Auth* related model.

This is where the models used by the authentication stack are defined.

It's perfectly fine to re-use this definition in the tg-sample-contacts application,
though.

"""
import os
from datetime import datetime
from hashlib import sha256
__all__ = ['User', 'Group', 'Permission']

from sqlalchemy import Table, ForeignKey, Column
from sqlalchemy.types import Unicode, Integer, DateTime
from sqlalchemy.orm import relation, synonym, relationship

from zope.sqlalchemy import ZopeTransactionExtension
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

maker = sessionmaker(autoflush=True, autocommit=False,
                     extension=ZopeTransactionExtension())
DBSession = scoped_session(maker)
DeclarativeBase = declarative_base()
metadata = DeclarativeBase.metadata


def init_model(engine):
    """Call me before using any of the tables or classes in the model."""
    DBSession.configure(bind=engine)
    return DBSession


# This is the association table for the many-to-many relationship between
# groups and permissions.
group_permission_table = Table('tg_group_permission', metadata,
                               Column('group_id', Integer,
                                      ForeignKey('tg_group.group_id',
                                                 onupdate="CASCADE",
                                                 ondelete="CASCADE"),
                                      primary_key=True),
                               Column('permission_id', Integer,
                                      ForeignKey('tg_permission.permission_id',
                                                 onupdate="CASCADE",
                                                 ondelete="CASCADE"),
                                      primary_key=True))


# This is the association table for the many-to-many relationship between
# groups and members - this is, the memberships.
user_group_table = Table('tg_user_group', metadata,
                         Column('user_id', Integer,
                                ForeignKey('tg_user.user_id',
                                           onupdate="CASCADE",
                                           ondelete="CASCADE"),
                                primary_key=True),
                         Column('group_id', Integer,
                                ForeignKey('tg_group.group_id',
                                           onupdate="CASCADE",
                                           ondelete="CASCADE"),
                                primary_key=True))


class Group(DeclarativeBase):
    """
    Group definition

    Only the ``group_name`` column is required.

    """

    __tablename__ = 'tg_group'

    group_id = Column(Integer, autoincrement=True, primary_key=True)
    group_name = Column(Unicode(16), unique=True, nullable=False)
    display_name = Column(Unicode(255))
    created = Column(DateTime, default=datetime.now)
    users = relation('User', secondary=user_group_table, backref='groups')

    def __repr__(self):
        return '<Group: name=%s>' % repr(self.group_name)

    def __unicode__(self):
        return self.group_name


class User(DeclarativeBase):
    """
    User definition.

    This is the user definition used by :mod:`repoze.who`, which requires at
    least the ``user_name`` column.

    """
    __tablename__ = 'tg_user'

    user_id = Column(Integer, autoincrement=True, primary_key=True)
    user_name = Column(Unicode(16), unique=True, nullable=False)
    email_address = Column(Unicode(255), unique=True, nullable=False)
    display_name = Column(Unicode(255))
    _password = Column('password', Unicode(128))
    created = Column(DateTime, default=datetime.now)

    def __repr__(self):
        return '<User: name=%s, email=%s, display=%s>' % (
            repr(self.user_name),
            repr(self.email_address),
            repr(self.display_name)
        )

    def __unicode__(self):
        return self.display_name or self.user_name

    @property
    def permissions(self):
        """Return a set with all permissions granted to the user."""
        perms = set()
        for g in self.groups:
            perms = perms | set(g.permissions)
        return perms

    @classmethod
    def by_email_address(cls, email):
        """Return the user object whose email address is ``email``."""
        return DBSession.query(cls).filter_by(email_address=email).first()

    @classmethod
    def by_user_name(cls, username):
        """Return the user object whose user name is ``username``."""
        return DBSession.query(cls).filter_by(user_name=username).first()

    @classmethod
    def _hash_password(cls, password):
        salt = sha256()
        salt.update(os.urandom(60))
        salt = salt.hexdigest()

        hash = sha256()
        # Make sure password is a str because we cannot hash unicode objects
        hash.update((password + salt).encode('utf-8'))
        hash = hash.hexdigest()

        password = salt + hash


        return password

    def _set_password(self, password):
        """Hash ``password`` on the fly and store its hashed version."""
        self._password = self._hash_password(password)

    def _get_password(self):
        """Return the hashed version of the password."""
        return self._password

    password = synonym('_password', descriptor=property(_get_password,
                                                        _set_password))

    def validate_password(self, password):
        """
        Check the password against existing credentials.

        :param password: the password that was provided by the user to
            try and authenticate. This is the clear text version that we will
            need to match against the hashed one in the database.
        :type password: unicode object.
        :return: Whether the password is valid.
        :rtype: bool

        """
        hash = sha256()
        hash.update((password + self.password[:64]).encode('utf-8'))
        return self.password[64:] == hash.hexdigest()


class Permission(DeclarativeBase):
    """
    Permission definition.

    Only the ``permission_name`` column is required.

    """

    __tablename__ = 'tg_permission'

    permission_id = Column(Integer, autoincrement=True, primary_key=True)
    permission_name = Column(Unicode(63), unique=True, nullable=False)
    description = Column(Unicode(255))

    groups = relation(Group, secondary=group_permission_table,
                      backref='permissions')

    def __repr__(self):
        return '<Permission: name=%s>' % repr(self.permission_name)

    def __unicode__(self):
        return self.permission_name



import bson
from webtest import TestApp
import transaction

from tg import AppConfig
from tg.configuration import milestones
from tg.configuration.auth import TGAuthMetadata
from tgext.pluggable import plug, app_model


class FakeAppPackage(object):
    __file__ = __file__
    __name__ = 'tests'

    class lib(object):
        class helpers(object):
            pass
        helpers = helpers()

        class app_globals(object):
            class Globals():
                pass
        app_globals = app_globals()

    class websetup(object):
        def bootstrap(*args, **kwargs):
            pass


class TestAuthMetadata(TGAuthMetadata):
    def authenticate(self, environ, identity):
        return 'user'

    def get_user(self, identity, userid):
        if userid:
            from tg.util import Bunch
            return Bunch(display_name='Fake Manager',
                         user_name='fake_manager',
                         email_address='fake@manager.it',
                         groups=[])

    def get_groups(self, identity, userid):
        if userid:
            return ['managers']
            # return Bunch(group_name='managers',
            #              display_name='Managers',
            #              permissions=[])
        return []

    def get_permissions(self, identity, userid):
        if userid:
            return ['tgapppermissions']
        return []


def configure_app(using):
    # Simulate starting configuration process from scratch
    milestones._reset_all()

    app_cfg = AppConfig(minimal=True)
    app_cfg.renderers = ['kajiki']
    app_cfg.default_renderer = 'kajiki'
    app_cfg.use_dotted_templatenames = True
    app_cfg.package = FakeAppPackage()
    app_cfg.use_toscawidgets2 = True
    app_cfg.sa_auth.authmetadata = TestAuthMetadata()
    app_cfg['beaker.session.secret'] = app_cfg['session.secret'] = 'SECRET'
    app_cfg.auth_backend = 'ming'

    if using == 'sqlalchemy':
        from . import sqlamodel
        app_cfg.package.model = sqlamodel
        app_cfg.use_sqlalchemy = True
        app_cfg['sqlalchemy.url'] = 'sqlite://'
        app_cfg.use_transaction_manager = True
        app_cfg.SQLASession = app_cfg.package.model.DBSession
    elif using == 'ming':
        from . import mingmodel
        app_cfg.package.model = mingmodel
        app_cfg.use_ming = True
        app_cfg['ming.url'] = 'mim:///fakedb'
        app_cfg.MingSession = app_cfg.package.model.DBSession
    else:
        raise ValueError('Unsupported backend')

    app_cfg.model = app_cfg.package.model
    app_cfg.DBSession = app_cfg.package.model.DBSession

    plug(app_cfg, 'tgapppermissions', plug_bootstrap=True)
    return app_cfg


def create_app(app_config, auth=False):
    app = app_config.make_wsgi_app(skip_authentication=True)

    if auth:
        app = TestApp(app, extra_environ=dict(REMOTE_USER='user'))
    else:
        app = TestApp(app)

    app.get('/non_existing_url_force_app_config_update', status=404)

    if app_config.get('use_ming'):
        datastore = app_config.DBSession.bind
        try:
            # On MIM drop all data
            datastore.conn.drop_all()
        except TypeError:
            # On MongoDB drop database
            datastore.conn.drop_database(datastore.db)
    elif app_config.get('use_sqlalchemy'):
        engine = app_config.DBSession.bind
        model = app_config['package'].model
        model.metadata.drop_all(engine)
        model.metadata.create_all(engine)
    else:
        raise ValueError('Unsupported backend')

    return app


def flush_db_changes():
    app_model.DBSession.flush()
    transaction.commit()



from formencode.validators import UnicodeString
from tg.i18n import lazy_ugettext as l_
from tw2.forms.widgets import Form, TextField, TextArea, SubmitButton, MultipleSelectField, EmailField
from tw2.core import Deferred
from axf.bootstrap import BootstrapFormLayout
from tgapppermissions.lib import helpers as h


class KajikiBootstrapFormLayout(BootstrapFormLayout):
    inline_engine_name = 'kajiki'


class NewPermission(Form):
    class child(KajikiBootstrapFormLayout):
        permission_name = TextField(label=l_('Permission Name'), css_class='form-control',
                         validator=UnicodeString(min=3))

        description = TextArea(label=l_('Permission Description'), rows=10, css_class='form-control',
                               validator=UnicodeString(min=3))

        groups = MultipleSelectField(label=l_('Permission Groups'), css_class="form-control",
                                     options=Deferred(h.query_groups))

    submit = SubmitButton(css_class='btn btn-primary pull-right', value=l_('Create'))


class EditPermission(NewPermission):
    submit = SubmitButton(css_class='btn btn-primary pull-right', value=l_('Edit'))


class EditUser(Form):
    class child(KajikiBootstrapFormLayout):
        display_name = TextField(label=l_('Display Name'),
                                 css_class='form-control',
                                 validator=UnicodeString(min=3, not_empty=False),
                                 attrs=dict(disabled='disabled'))
        user_name = TextField(label=l_('User Name'),
                              css_class='form-control',
                              validator=UnicodeString(min=3, not_empty=False),
                              attrs=dict(disabled='disabled'))
        email_address = EmailField(label=l_('Email Address'),
                                   css_class='form-control',
                                   attrs=dict(disabled='disabled'))
        created = TextField(label=l_('Created'),
                            css_class='form-control',
                            attrs=dict(disabled='disabled'))
        groups = MultipleSelectField(label=l_('Groups'),
                                     css_class='form-control',
                                     options=Deferred(h.query_groups))

    submit = SubmitButton(css_class='btn btn-primary pull-right', value=l_('Edit'))



from tgapppermissions import model
from tgext.pluggable import app_model


def query_groups():
    _, groups = model.provider.query(app_model.Group)
    primary_field = model.provider.get_primary_field(app_model.Group)
    return [(getattr(group, primary_field), group.display_name) for group in groups]



# -*- coding: utf-8 -*-

"""WebHelpers used in tgapp-permissions."""

from tgext.pluggable import plug_url
from tgapppermissions import model


def get_primary_field(_model):
    return model.provider.get_primary_field(
        model.provider.get_entity(_model)
    )


# waiting for a new version of tgext.pluggable
try:
    from tgext.pluggable.utils import instance_primary_key
except ImportError:  # when tgext.pluggable is not updated
    def instance_primary_key(instance, as_string=False):
        """Returns the value of the primary key of the instance"""
        from tgext.pluggable.utils import primary_key
        p = getattr(instance, primary_key(instance.__class__).name)
        return p if not as_string else str(p)



import tg
from tgext.pluggable import app_model
from tgapppermissions import model
from .base import configure_app, create_app, flush_db_changes
import re
from webtest import AppError
from datetime import datetime
from tgapppermissions.helpers import get_primary_field


find_urls = re.compile('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')


class RegistrationControllerTests(object):
    def setup(self):
        self.app = create_app(self.app_config, False)

    def test_index(self):
        resp = self.app.get('/')
        assert 'HELLO' in resp.text

    def test_tgapppermissions_index(self):
        resp = self.app.get('/tgapppermissions', extra_environ={'REMOTE_USER': 'manager'})

        assert '/users' in resp.text, resp
        assert '/new_permission' in resp.text, resp

    def test_tgapppermissions_auth(self):
        try:
            self.app.get('/tgapppermissions')
        except AppError as e:
            assert '401' in str(e)

    def test_create_permission(self):
        g_primary = get_primary_field(app_model.Group)

        g = model.provider.create(app_model.Group,
                                  {'group_name': 'editors', 'dispaly_name': 'Editors'})
        g_primary_value = getattr(g, g_primary)
        flush_db_changes()

        self.app.get(
            '/tgapppermissions/create_permission',
             params={'permission_name': 'pname',
                     'description': 'descr',
                     'groups': [g_primary_value]},
             extra_environ={'REMOTE_USER': 'manager'},
             status=302,
             )
        count, perms = model.provider.query(app_model.Permission,
                                            filters=dict(permission_name='pname'))

        assert count == 1
        perm = perms[0]
        assert 'pname' == perm.permission_name, perm.permission_name
        assert 'descr' == perm.description, perm.description
        assert g_primary_value == getattr(perm.groups[0], g_primary), perm.groups[0]

    def test_update_permission(self):
        g_primary = get_primary_field(app_model.Group)
        p_primary = get_primary_field(app_model.Permission)

        g1 = model.provider.create(app_model.Group,
                                   {'group_name': 'editors', 'dispaly_name': 'Editors'})
        g1_primary_value = getattr(g1, g_primary)
        g2 = model.provider.create(app_model.Group,
                                   {'group_name': 'users', 'dispaly_name': 'Users'})
        g2_primary_value = getattr(g2, g_primary)
        flush_db_changes()

        self.app.get(
            '/tgapppermissions/create_permission',
             params={'permission_name': 'pname',
                     'description': 'descr',
                     'groups': []},
             extra_environ={'REMOTE_USER': 'manager'},
             status=302,
             )
        count, perms = model.provider.query(app_model.Permission,
                                            filters=dict(permission_name='pname'))

        assert count == 1
        perm = perms[0]

        self.app.get('/tgapppermissions/update_permission/' + str(getattr(perm, p_primary)),
                     params={'permission_name': 'view',
                             'description': 'perm to view things',
                             'groups': [g1_primary_value, g2_primary_value]},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302,
                     )
        count, perms = model.provider.query(app_model.Permission,
                                            filters=dict(permission_name='view'))

        assert count == 1
        perm = perms[0]

        assert 'view' == perm.permission_name, perm.permission_name
        assert 'perm to view things' == perm.description, perm.description
        # assert perm.groups == [g1, g2], perm.groups
        _, groups = model.provider.query(app_model.Group)
        assert set(
            [getattr(g, g_primary) for g in perm.groups]
        ) == set(
            [getattr(g, g_primary) for g in groups]
        )

    def test_delete_permission(self):
        p_primary = get_primary_field(app_model.Permission)
        self.app.get(
            '/tgapppermissions/create_permission',
            params={'permission_name': 'pname',
                    'description': 'descr',
                    'groups': []},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        count, perms = model.provider.query(app_model.Permission)

        assert count == 1
        perm = perms[0]

        self.app.get('/tgapppermissions/delete_permission/' + str(getattr(perm, p_primary)),
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302)
        count, perms = model.provider.query(app_model.Permission)

        assert count == 0

    def test_new_permission_form(self):
        resp = self.app.get('/tgapppermissions/new_permission',
                            extra_environ={'REMOTE_USER': 'manager'})

        assert 'name="permission_name"' in resp.text, resp
        assert 'name="description"' in resp.text, resp
        assert 'name="groups"' in resp.text, resp
        assert '/create_permission' in resp.text, resp

    def test_edit_permission_form_without_perm_id(self):
        self.app.get('/tgapppermissions/edit_permission',
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=404)

    def test_edit_permission_form(self):
        p_primary = get_primary_field(app_model.Permission)
        self.app.get(
            '/tgapppermissions/create_permission',
            params={'permission_name': 'pname',
                    'description': 'descr',
                    'groups': []},
            extra_environ={'REMOTE_USER': 'manager'},
            status=302,
        )
        count, perms = model.provider.query(app_model.Permission)

        assert count == 1
        perm = perms[0]

        resp = self.app.get('/tgapppermissions/edit_permission/' + str(getattr(perm, p_primary)),
                            extra_environ={'REMOTE_USER': 'manager'})

        assert 'name="permission_name"' in resp.text, resp
        assert 'name="description"' in resp.text, resp
        assert 'name="groups' in resp.text, resp
        assert '/update_permission' in resp.text, resp

    def test_users(self):
        g_primary = get_primary_field(app_model.Group)
        u_primary = get_primary_field(app_model.User)

        g1 = model.provider.create(app_model.Group,
                                   {'group_name': 'editors', 'display_name': 'Editors'})
        g1_primary_value = getattr(g1, g_primary)
        u1 = model.provider.create(app_model.User,
                                   {'user_name': 'alpha',
                                    'display_name': 'Alpha',
                                    'email_address': 'email@email.email',
                                    'created': datetime.utcnow(),
                                    'groups': [g1_primary_value]})
        u1_primary_value = getattr(u1, u_primary)
        flush_db_changes()

        resp = self.app.get('/tgapppermissions/users',
                            extra_environ={'REMOTE_USER': 'manager'})

        assert '/edit_user/' + str(u1_primary_value) in resp.text, resp
        assert 'alpha' in resp.text, resp
        assert 'Alpha' in resp.text, resp
        assert 'email@email.email' in resp.text, resp
        assert 'Editors' in resp.text, resp

    def test_users_search(self):
        model.provider.create(app_model.User,
                              {'user_name': 'alpha',
                               'display_name': 'Alpha',
                               'email_address': 'email@email.email',
                               'created': datetime.utcnow()})
        model.provider.create(app_model.User,
                              {'user_name': 'ciao',
                               'display_name': 'Ciao',
                               'email_address': 'ciao@email.email',
                               'created': datetime.utcnow()})
        model.provider.create(app_model.User,
                              {'user_name': 'miao',
                               'display_name': 'Miao',
                               'email_address': 'miao@email.email',
                               'created': datetime.utcnow()})
        flush_db_changes()

        resp = self.app.get('/tgapppermissions/users/?search_by=user_name&search_value=iao',
                            extra_environ={'REMOTE_USER': 'manager'})

        assert 'Alpha' not in resp.text, resp
        assert 'Ciao' in resp.text, resp
        assert 'Miao' in resp.text, resp

    def test_edit_user(self):
        g_primary = get_primary_field(app_model.Group)
        u_primary = get_primary_field(app_model.User)

        u1 = model.provider.create(app_model.User,
                                   {'user_name': 'alpha',
                                    'display_name': 'Alpha',
                                    'email_address': 'email@email.email',
                                    'created': datetime.utcnow()})
        u1_primary_value = getattr(u1, u_primary)
        g1 = model.provider.create(app_model.Group,
                                   {'group_name': 'editors', 'display_name': 'Editors'})
        g1_primary_value = getattr(g1, g_primary)
        flush_db_changes()

        resp = self.app.get('/tgapppermissions/edit_user/' + str(u1_primary_value),
                            extra_environ={'REMOTE_USER': 'manager'},)
                            # {'user_name': 'beta',
                            #  'display_name': 'Beta',
                            #  'email_address': 'email2@email2.email2',
                            #  'created': datetime.utcnow(),
                            #  'groups': [g1_primary_value]})

        assert '/update_user/' + str(u1_primary_value) in resp.text, resp
        assert 'alpha' in resp.text, resp
        assert 'Alpha' in resp.text, resp
        assert 'email@email.email' in resp.text, resp

    def test_update_user(self):
        g_primary = get_primary_field(app_model.Group)
        u_primary = get_primary_field(app_model.User)

        u1 = model.provider.create(app_model.User,
                                   {'user_name': 'alpha',
                                    'display_name': 'Alpha',
                                    'email_address': 'email@email.email',
                                    'created': datetime.utcnow()})
        u1_primary_value = getattr(u1, u_primary)
        g1 = model.provider.create(app_model.Group,
                                   {'group_name': 'editors', 'display_name': 'Editors'})
        g1_primary_value = getattr(g1, g_primary)
        flush_db_changes()

        self.app.get('/tgapppermissions/update_user/' + str(u1_primary_value),
                     {'user_name': 'beta',
                      'display_name': 'Beta',
                      'email_address': 'email2@email2.email2',
                      'created': datetime.utcnow(),
                      'groups': [g1_primary_value]},
                     extra_environ={'REMOTE_USER': 'manager'},
                     status=302)

        count, groups = model.provider.query(app_model.Group)
        count, users = model.provider.query(app_model.User)

        assert count == 1
        usr = users[0]

        assert u1_primary_value == getattr(usr, u_primary), usr
        # update_user currently updates ONLY the groups of the user
        # assert 'beta' == usr.user_name, usr
        # assert 'Beta' == usr.display_name, usr
        # assert 'email2@email2.email2' == usr.email_address
        assert set(
            [getattr(g, g_primary) for g in usr.groups]
        ) == set(
            [getattr(g, g_primary) for g in groups]
        )


class TestRegistrationControllerSQLA(RegistrationControllerTests):
    @classmethod
    def setupClass(cls):
        cls.app_config = configure_app('sqlalchemy')


class TestRegistrationControllerMing(RegistrationControllerTests):
    @classmethod
    def setupClass(cls):
        cls.app_config = configure_app('ming')
